@@ -36,6 +36,7 @@
 #endif
 
 #include <gdbus.h>
+#include <connman/ipaddress.h>
 
 #include "connman.h"
 
@@ -89,171 +90,14 @@ struct connman_ipdevice {
 
 static GHashTable *ipdevice_hash = NULL;
 static GList *ipconfig_list = NULL;
+static connman_bool_t is_ipv6_supported = FALSE;
 
-struct connman_ipaddress *connman_ipaddress_alloc(int family)
+void __connman_ipconfig_clear_address(struct connman_ipconfig *ipconfig)
 {
-	struct connman_ipaddress *ipaddress;
-
-	ipaddress = g_try_new0(struct connman_ipaddress, 1);
-	if (ipaddress == NULL)
-		return NULL;
-
-	ipaddress->family = family;
-	ipaddress->prefixlen = 0;
-	ipaddress->local = NULL;
-	ipaddress->peer = NULL;
-	ipaddress->broadcast = NULL;
-	ipaddress->gateway = NULL;
-
-	return ipaddress;
-}
-
-void connman_ipaddress_free(struct connman_ipaddress *ipaddress)
-{
-	if (ipaddress == NULL)
+	if (ipconfig == NULL)
 		return;
 
-	g_free(ipaddress->broadcast);
-	g_free(ipaddress->peer);
-	g_free(ipaddress->local);
-	g_free(ipaddress->gateway);
-	g_free(ipaddress);
-}
-
-unsigned char __connman_ipconfig_netmask_prefix_len(const char *netmask)
-{
-	unsigned char bits;
-	in_addr_t mask;
-	in_addr_t host;
-
-	if (netmask == NULL)
-		return 32;
-
-	mask = inet_network(netmask);
-	host = ~mask;
-
-	/* a valid netmask must be 2^n - 1 */
-	if ((host & (host + 1)) != 0)
-		return -1;
-
-	bits = 0;
-	for (; mask; mask <<= 1)
-		++bits;
-
-	return bits;
-}
-
-static gboolean check_ipv6_address(const char *address)
-{
-	unsigned char buf[sizeof(struct in6_addr)];
-	int err;
-
-	if (address == NULL)
-		return FALSE;
-
-	err = inet_pton(AF_INET6, address, buf);
-	if (err > 0)
-		return TRUE;
-
-	return FALSE;
-}
-
-int connman_ipaddress_set_ipv6(struct connman_ipaddress *ipaddress,
-				const char *address,
-				unsigned char prefix_length,
-				const char *gateway)
-{
-	if (ipaddress == NULL)
-		return -EINVAL;
-
-	if (check_ipv6_address(address) == FALSE)
-		return -EINVAL;
-
-	DBG("prefix_len %d address %s gateway %s",
-			prefix_length, address, gateway);
-
-	ipaddress->family = AF_INET6;
-
-	ipaddress->prefixlen = prefix_length;
-
-	g_free(ipaddress->local);
-	ipaddress->local = g_strdup(address);
-
-	g_free(ipaddress->gateway);
-	ipaddress->gateway = g_strdup(gateway);
-
-	return 0;
-}
-
-int connman_ipaddress_set_ipv4(struct connman_ipaddress *ipaddress,
-		const char *address, const char *netmask, const char *gateway)
-{
-	if (ipaddress == NULL)
-		return -EINVAL;
-
-	ipaddress->family = AF_INET;
-
-	ipaddress->prefixlen = __connman_ipconfig_netmask_prefix_len(netmask);
-
-	g_free(ipaddress->local);
-	ipaddress->local = g_strdup(address);
-
-	g_free(ipaddress->gateway);
-	ipaddress->gateway = g_strdup(gateway);
-
-	return 0;
-}
-
-void connman_ipaddress_set_peer(struct connman_ipaddress *ipaddress,
-				const char *peer)
-{
-	if (ipaddress == NULL)
-		return;
-
-	g_free(ipaddress->peer);
-	ipaddress->peer = g_strdup(peer);
-}
-
-void connman_ipaddress_clear(struct connman_ipaddress *ipaddress)
-{
-	if (ipaddress == NULL)
-		return;
-
-	ipaddress->prefixlen = 0;
-
-	g_free(ipaddress->local);
-	ipaddress->local = NULL;
-
-	g_free(ipaddress->peer);
-	ipaddress->peer = NULL;
-
-	g_free(ipaddress->broadcast);
-	ipaddress->broadcast = NULL;
-
-	g_free(ipaddress->gateway);
-	ipaddress->gateway = NULL;
-}
-
-void connman_ipaddress_copy(struct connman_ipaddress *ipaddress,
-					struct connman_ipaddress *source)
-{
-	if (ipaddress == NULL || source == NULL)
-		return;
-
-	ipaddress->family = source->family;
-	ipaddress->prefixlen = source->prefixlen;
-
-	g_free(ipaddress->local);
-	ipaddress->local = g_strdup(source->local);
-
-	g_free(ipaddress->peer);
-	ipaddress->peer = g_strdup(source->peer);
-
-	g_free(ipaddress->broadcast);
-	ipaddress->broadcast = g_strdup(source->broadcast);
-
-	g_free(ipaddress->gateway);
-	ipaddress->gateway = g_strdup(source->gateway);
+	connman_ipaddress_clear(ipconfig->address);
 }
 
 static void free_address_list(struct connman_ipdevice *ipdevice)
@@ -547,15 +391,6 @@ static void __connman_ipconfig_lower_up(struct connman_ipdevice *ipdevice)
 {
 	DBG("ipconfig ipv4 %p ipv6 %p", ipdevice->config_ipv4,
 					ipdevice->config_ipv6);
-
-	if (ipdevice->config_ipv6 != NULL &&
-			ipdevice->config_ipv6->enabled == TRUE)
-		return;
-
-	if (__connman_device_isfiltered(ipdevice->ifname) == FALSE) {
-		ipdevice->ipv6_enabled = get_ipv6_state(ipdevice->ifname);
-		set_ipv6_state(ipdevice->ifname, FALSE);
-	}
 }
 
 static void __connman_ipconfig_lower_down(struct connman_ipdevice *ipdevice)
@@ -1266,7 +1101,11 @@ static struct connman_ipconfig *create_ipv6config(int index)
 	ipv6config->index = index;
 	ipv6config->enabled = FALSE;
 	ipv6config->type = CONNMAN_IPCONFIG_TYPE_IPV6;
-	ipv6config->method = CONNMAN_IPCONFIG_METHOD_AUTO;
+
+	if (is_ipv6_supported == FALSE)
+		ipv6config->method = CONNMAN_IPCONFIG_METHOD_OFF;
+	else
+		ipv6config->method = CONNMAN_IPCONFIG_METHOD_AUTO;
 
 	ipdevice = g_hash_table_lookup(ipdevice_hash, GINT_TO_POINTER(index));
 	if (ipdevice != NULL)
@@ -1280,7 +1119,8 @@ static struct connman_ipconfig *create_ipv6config(int index)
 
 	ipv6config->system = connman_ipaddress_alloc(AF_INET6);
 
-	DBG("ipconfig %p", ipv6config);
+	DBG("ipconfig %p method %s", ipv6config,
+		__connman_ipconfig_method2string(ipv6config->method));
 
 	return ipv6config;
 }
@@ -1716,7 +1556,6 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 		if (ipdevice->config_ipv6 == ipconfig)
 			return -EALREADY;
 		type = CONNMAN_IPCONFIG_TYPE_IPV6;
-		enable_ipv6(ipconfig);
 	} else
 		return -EINVAL;
 
@@ -1744,9 +1583,11 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
 		ipdevice->config_ipv4 = __connman_ipconfig_ref(ipconfig);
-	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
+	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6) {
 		ipdevice->config_ipv6 = __connman_ipconfig_ref(ipconfig);
 
+		enable_ipv6(ipdevice->config_ipv6);
+	}
 	ipconfig_list = g_list_append(ipconfig_list, ipconfig);
 
 	if (ipdevice->flags & IFF_UP)
@@ -2089,7 +1930,7 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 {
 	enum connman_ipconfig_method method = CONNMAN_IPCONFIG_METHOD_UNKNOWN;
 	const char *address = NULL, *netmask = NULL, *gateway = NULL,
-		*prefix_length_string = NULL, *privacy_string = NULL;
+		*privacy_string = NULL;
 	int prefix_length = 0, privacy = 0;
 	DBusMessageIter dict;
 
@@ -2134,13 +1975,11 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 
 			dbus_message_iter_get_basic(&value, &address);
 		} else if (g_str_equal(key, "PrefixLength") == TRUE) {
-			if (type != DBUS_TYPE_STRING)
+			if (type != DBUS_TYPE_BYTE)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&value,
-							&prefix_length_string);
+			dbus_message_iter_get_basic(&value, &prefix_length);
 
-			prefix_length = atoi(prefix_length_string);
 			if (prefix_length < 0 || prefix_length > 128)
 				return -EINVAL;
 		} else if (g_str_equal(key, "Netmask") == TRUE) {
@@ -2406,6 +2245,8 @@ int __connman_ipconfig_init(void)
 	ipdevice_hash = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 							NULL, free_ipdevice);
 
+	is_ipv6_supported = connman_inet_is_ipv6_supported();
+
 	return 0;
 }
 
