@@ -55,7 +55,7 @@ struct connman_technology {
 	enum connman_service_type type;
 	char *path;
 	GSList *device_list;
-	int enabled;
+	connman_bool_t enabled;
 	char *regdom;
 	connman_bool_t connected;
 
@@ -65,14 +65,15 @@ struct connman_technology {
 
 	connman_bool_t enable_persistent; /* Save the tech state */
 
-	struct connman_technology_driver *driver;
-	void *driver_data;
+	GSList *driver_list;
 
 	DBusMessage *pending_reply;
 	guint pending_timeout;
 
 	GSList *scan_pending;
 
+	connman_bool_t rfkill_driven;
+	connman_bool_t softblocked;
 	connman_bool_t hardblocked;
 	connman_bool_t dbus_registered;
 };
@@ -117,14 +118,11 @@ int connman_technology_driver_register(struct connman_technology_driver *driver)
 	driver_list = g_slist_insert_sorted(driver_list, driver,
 							compare_priority);
 
-	if (techless_device_list == NULL)
-		goto check_rfkill;
-
 	/*
 	 * Check for technology less devices if this driver
 	 * can service any of them.
 	*/
-	for (list = techless_device_list; list; list = list->next) {
+	for (list = techless_device_list; list != NULL; list = list->next) {
 		device = list->data;
 
 		type = __connman_device_get_service_type(device);
@@ -137,7 +135,6 @@ int connman_technology_driver_register(struct connman_technology_driver *driver)
 		__connman_technology_add_device(device);
 	}
 
-check_rfkill:
 	/* Check for orphaned rfkill switches. */
 	g_hash_table_foreach(rfkill_list, rfkill_check,
 					GINT_TO_POINTER(driver->type));
@@ -153,20 +150,30 @@ check_rfkill:
  */
 void connman_technology_driver_unregister(struct connman_technology_driver *driver)
 {
-	GSList *list;
+	GSList *list, *tech_drivers;
 	struct connman_technology *technology;
+	struct connman_technology_driver *current;
 
 	DBG("Unregistering driver %p name %s", driver, driver->name);
 
 	for (list = technology_list; list; list = list->next) {
 		technology = list->data;
 
-		if (technology->driver == NULL)
-			continue;
+		for (tech_drivers = technology->driver_list;
+		     tech_drivers != NULL;
+		     tech_drivers = g_slist_next(tech_drivers)) {
 
-		if (technology->type == driver->type) {
-			technology->driver->remove(technology);
-			technology->driver = NULL;
+			current = tech_drivers->data;
+			if (driver != current)
+				continue;
+
+			if (driver->remove != NULL)
+				driver->remove(technology);
+
+			technology->driver_list =
+				g_slist_remove(technology->driver_list, driver);
+
+			break;
 		}
 	}
 
@@ -212,14 +219,17 @@ void connman_technology_tethering_notify(struct connman_technology *technology,
 static int set_tethering(struct connman_technology *technology,
 				connman_bool_t enabled)
 {
+	int result = -EOPNOTSUPP;
+	int err;
 	const char *ident, *passphrase, *bridge;
+	GSList *tech_drivers;
 
 	ident = technology->tethering_ident;
 	passphrase = technology->tethering_passphrase;
 
-	if (technology->driver == NULL ||
-			technology->driver->set_tethering == NULL)
-		return -EOPNOTSUPP;
+	__sync_synchronize();
+	if (technology->enabled == FALSE)
+		return -EACCES;
 
 	bridge = __connman_tethering_get_bridge();
 	if (bridge == NULL)
@@ -229,8 +239,26 @@ static int set_tethering(struct connman_technology *technology,
 	    (ident == NULL || passphrase == NULL))
 		return -EINVAL;
 
-	return technology->driver->set_tethering(technology, ident, passphrase,
-							bridge, enabled);
+	for (tech_drivers = technology->driver_list; tech_drivers != NULL;
+	     tech_drivers = g_slist_next(tech_drivers)) {
+		struct connman_technology_driver *driver = tech_drivers->data;
+
+		if (driver == NULL || driver->set_tethering == NULL)
+			continue;
+
+		err = driver->set_tethering(technology, ident, passphrase,
+				bridge, enabled);
+
+		if (result == -EINPROGRESS)
+			continue;
+
+		if (err == -EINPROGRESS || err == 0) {
+			result = err;
+			continue;
+		}
+	}
+
+	return result;
 }
 
 void connman_technology_regdom_notify(struct connman_technology *technology,
@@ -264,18 +292,23 @@ static int set_regdom_by_device(struct connman_technology *technology,
 
 int connman_technology_set_regdom(const char *alpha2)
 {
-	GSList *list;
+	GSList *list, *tech_drivers;
 
 	for (list = technology_list; list; list = list->next) {
 		struct connman_technology *technology = list->data;
 
 		if (set_regdom_by_device(technology, alpha2) != 0) {
-			if (technology->driver == NULL)
-				continue;
 
-			if (technology->driver->set_regdom != NULL)
-				technology->driver->set_regdom(technology,
-								alpha2);
+			for (tech_drivers = technology->driver_list;
+			     tech_drivers != NULL;
+			     tech_drivers = g_slist_next(tech_drivers)) {
+
+				struct connman_technology_driver *driver =
+					tech_drivers->data;
+
+				if (driver->set_regdom != NULL)
+					driver->set_regdom(technology, alpha2);
+			}
 		}
 	}
 
@@ -302,8 +335,6 @@ static const char *get_name(enum connman_service_type type)
 		return "Wired";
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		return "WiFi";
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-		return "WiMAX";
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 		return "Bluetooth";
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
@@ -452,7 +483,6 @@ static void append_properties(DBusMessageIter *iter,
 {
 	DBusMessageIter dict;
 	const char *str;
-	connman_bool_t powered;
 
 	connman_dbus_dict_open(iter, &dict);
 
@@ -467,12 +497,9 @@ static void append_properties(DBusMessageIter *iter,
 						DBUS_TYPE_STRING, &str);
 
 	__sync_synchronize();
-	if (technology->enabled > 0)
-		powered = TRUE;
-	else
-		powered = FALSE;
 	connman_dbus_dict_append_basic(&dict, "Powered",
-					DBUS_TYPE_BOOLEAN, &powered);
+					DBUS_TYPE_BOOLEAN,
+					&technology->enabled);
 
 	connman_dbus_dict_append_basic(&dict, "Connected",
 					DBUS_TYPE_BOOLEAN,
@@ -484,13 +511,13 @@ static void append_properties(DBusMessageIter *iter,
 
 	if (technology->tethering_ident != NULL)
 		connman_dbus_dict_append_basic(&dict, "TetheringIdentifier",
-						DBUS_TYPE_STRING,
-						&technology->tethering_ident);
+					DBUS_TYPE_STRING,
+					&technology->tethering_ident);
 
 	if (technology->tethering_passphrase != NULL)
 		connman_dbus_dict_append_basic(&dict, "TetheringPassphrase",
-						DBUS_TYPE_STRING,
-						&technology->tethering_passphrase);
+					DBUS_TYPE_STRING,
+					&technology->tethering_passphrase);
 
 	connman_dbus_dict_close(iter, &dict);
 }
@@ -548,7 +575,8 @@ void __connman_technology_list_struct(DBusMessageIter *array)
 		struct connman_technology *technology = list->data;
 
 		if (technology->path == NULL ||
-					technology->hardblocked == TRUE)
+				(technology->rfkill_driven == TRUE &&
+				 technology->hardblocked == TRUE))
 			continue;
 
 		dbus_message_iter_open_container(array, DBUS_TYPE_STRUCT,
@@ -579,72 +607,70 @@ static gboolean technology_pending_reply(gpointer user_data)
 	return FALSE;
 }
 
-static int technology_enable(struct connman_technology *technology,
-						connman_bool_t hardblock)
+static int technology_affect_devices(struct connman_technology *technology,
+						connman_bool_t enable_device)
 {
 	GSList *list;
 	int err = 0;
 
+	for (list = technology->device_list; list; list = list->next) {
+		struct connman_device *device = list->data;
+
+		if (enable_device == TRUE)
+			err = __connman_device_enable(device);
+		else
+			err = __connman_device_disable(device);
+	}
+
+	return err;
+}
+
+static int technology_enable(struct connman_technology *technology)
+{
+	int err = 0;
+	int err_dev;
+
 	DBG("technology %p enable", technology);
 
 	__sync_synchronize();
-	if (technology->enabled > 0) {
-		err = -EALREADY;
-		goto done;
-	}
+	if (technology->enabled == TRUE)
+		return -EALREADY;
 
-	if (technology->pending_reply != NULL) {
-		err = -EBUSY;
-		goto done;
-	}
+	if (technology->pending_reply != NULL)
+		return -EBUSY;
 
-	if (hardblock == TRUE && technology->enable_persistent == FALSE)
-		goto done;
+	if (technology->rfkill_driven == TRUE)
+		err = __connman_rfkill_block(technology->type, FALSE);
 
-	__connman_rfkill_block(technology->type, FALSE);
+	err_dev = technology_affect_devices(technology, TRUE);
 
-	for (list = technology->device_list; list; list = list->next) {
-		struct connman_device *device = list->data;
+	if (technology->rfkill_driven == FALSE)
+		err = err_dev;
 
-		err = __connman_device_enable(device);
-	}
-
-done:
 	return err;
 }
 
-static int technology_disable(struct connman_technology *technology,
-						connman_bool_t hardblock)
+static int technology_disable(struct connman_technology *technology)
 {
-	GSList *list;
-	int err = 0;
+	int err;
 
 	DBG("technology %p disable", technology);
 
 	__sync_synchronize();
-	if (technology->enabled == 0) {
-		err = -EALREADY;
-		goto done;
-	}
+	if (technology->enabled == FALSE)
+		return -EALREADY;
 
-	if (technology->pending_reply != NULL) {
-		err = -EBUSY;
-		goto done;
-	}
+	if (technology->pending_reply != NULL)
+		return -EBUSY;
 
 	if (technology->tethering == TRUE)
 		set_tethering(technology, FALSE);
 
-	if (hardblock == FALSE)
-		__connman_rfkill_block(technology->type, TRUE);
+	err = technology_affect_devices(technology, FALSE);
 
-	for (list = technology->device_list; list; list = list->next) {
-		struct connman_device *device = list->data;
+	if (technology->rfkill_driven == TRUE)
+		err = __connman_rfkill_block(technology->type, TRUE);
 
-		err = __connman_device_disable(device);
-	}
-
-done:
 	return err;
 }
 
@@ -654,15 +680,15 @@ static DBusMessage *set_powered(struct connman_technology *technology,
 	DBusMessage *reply = NULL;
 	int err = 0;
 
-	if (technology->hardblocked == TRUE) {
+	if (technology->rfkill_driven && technology->hardblocked == TRUE) {
 		err = -EACCES;
 		goto make_reply;
 	}
 
 	if (powered == TRUE)
-		err = technology_enable(technology, FALSE);
+		err = technology_enable(technology);
 	else
-		err = technology_disable(technology, FALSE);
+		err = technology_disable(technology);
 
 	if (err != -EBUSY) {
 		technology->enable_persistent = powered;
@@ -874,8 +900,10 @@ void __connman_technology_scan_stopped(struct connman_device *device)
 void __connman_technology_notify_regdom_by_device(struct connman_device *device,
 						int result, const char *alpha2)
 {
+	connman_bool_t regdom_set = FALSE;
 	struct connman_technology *technology;
 	enum connman_service_type type;
+	GSList *tech_drivers;
 
 	type = __connman_device_get_service_type(device);
 	technology = technology_find(type);
@@ -884,13 +912,22 @@ void __connman_technology_notify_regdom_by_device(struct connman_device *device,
 		return;
 
 	if (result < 0) {
-		if (technology->driver != NULL &&
-				technology->driver->set_regdom != NULL) {
-			technology->driver->set_regdom(technology, alpha2);
-			return;
+
+		for (tech_drivers = technology->driver_list;
+		     tech_drivers != NULL;
+		     tech_drivers = g_slist_next(tech_drivers)) {
+			struct connman_technology_driver *driver =
+				tech_drivers->data;
+
+			if (driver->set_regdom != NULL) {
+				driver->set_regdom(technology, alpha2);
+				regdom_set = TRUE;
+			}
+
 		}
 
-		alpha2 = NULL;
+		if (regdom_set == FALSE)
+			alpha2 = NULL;
 	}
 
 	connman_technology_regdom_notify(technology, alpha2);
@@ -935,7 +972,8 @@ static const GDBusSignalTable technology_signals[] = {
 static gboolean technology_dbus_register(struct connman_technology *technology)
 {
 	if (technology->dbus_registered == TRUE ||
-					technology->hardblocked == TRUE)
+				(technology->rfkill_driven == TRUE &&
+				 technology->hardblocked == TRUE))
 		return TRUE;
 
 	if (g_dbus_register_interface(connection, technology->path,
@@ -954,11 +992,11 @@ static gboolean technology_dbus_register(struct connman_technology *technology)
 
 static struct connman_technology *technology_get(enum connman_service_type type)
 {
+	GSList *tech_drivers = NULL;
+	struct connman_technology_driver *driver;
 	struct connman_technology *technology;
-	struct connman_technology_driver *driver = NULL;
 	const char *str;
 	GSList *list;
-	int err;
 
 	DBG("type %d", type);
 
@@ -976,14 +1014,14 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 	for (list = driver_list; list; list = list->next) {
 		driver = list->data;
 
-		if (driver->type == type)
-			break;
-		else
-			driver = NULL;
+		if (driver->type == type) {
+			DBG("technology %p driver %p", technology, driver);
+			tech_drivers = g_slist_append(tech_drivers, driver);
+		}
 	}
 
-	if (driver == NULL) {
-		DBG("No matching driver found for %s.",
+	if (tech_drivers == NULL) {
+		DBG("No matching drivers found for %s.",
 				__connman_service_type2string(type));
 		return NULL;
 	}
@@ -994,10 +1032,9 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 
 	technology->refcount = 1;
 
-	if (type == CONNMAN_SERVICE_TYPE_ETHERNET)
-		technology->hardblocked = FALSE;
-	else
-		technology->hardblocked = TRUE;
+	technology->rfkill_driven = FALSE;
+	technology->softblocked = FALSE;
+	technology->hardblocked = FALSE;
 
 	technology->type = type;
 	technology->path = g_strdup_printf("%s/technology/%s",
@@ -1016,10 +1053,15 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 
 	technology_list = g_slist_prepend(technology_list, technology);
 
-	technology->driver = driver;
-	err = driver->probe(technology);
-	if (err != 0)
-		DBG("Driver probe failed for technology %p", technology);
+	technology->driver_list = tech_drivers;
+
+	for (list = tech_drivers; list != NULL; list = g_slist_next(list)) {
+		driver = list->data;
+
+		if (driver->probe != NULL && driver->probe(technology) < 0)
+			DBG("Driver probe failed for technology %p",
+					technology);
+	}
 
 	DBG("technology %p", technology);
 
@@ -1047,9 +1089,17 @@ static void technology_put(struct connman_technology *technology)
 
 	reply_scan_pending(technology, -EINTR);
 
-	if (technology->driver) {
-		technology->driver->remove(technology);
-		technology->driver = NULL;
+	while (technology->driver_list != NULL) {
+		struct connman_technology_driver *driver;
+
+		driver = technology->driver_list->data;
+
+		if (driver->remove != NULL)
+			driver->remove(technology);
+
+		technology->driver_list =
+			g_slist_delete_link(technology->driver_list,
+					technology->driver_list);
 	}
 
 	technology_list = g_slist_remove(technology_list, technology);
@@ -1069,6 +1119,8 @@ void __connman_technology_add_interface(enum connman_service_type type,
 				int index, const char *name, const char *ident)
 {
 	struct connman_technology *technology;
+	GSList *tech_drivers;
+	struct connman_technology_driver *driver;
 
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
@@ -1076,7 +1128,6 @@ void __connman_technology_add_interface(enum connman_service_type type,
 		return;
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
 	case CONNMAN_SERVICE_TYPE_GPS:
@@ -1090,18 +1141,24 @@ void __connman_technology_add_interface(enum connman_service_type type,
 
 	technology = technology_find(type);
 
-	if (technology == NULL || technology->driver == NULL
-			|| technology->driver->add_interface == NULL)
+	if (technology == NULL)
 		return;
 
-	technology->driver->add_interface(technology,
-					index, name, ident);
+	for (tech_drivers = technology->driver_list; tech_drivers != NULL;
+	     tech_drivers = g_slist_next(tech_drivers)) {
+		driver = tech_drivers->data;
+
+		if(driver->add_interface != NULL)
+			driver->add_interface(technology, index, name, ident);
+	}
 }
 
 void __connman_technology_remove_interface(enum connman_service_type type,
 				int index, const char *name, const char *ident)
 {
 	struct connman_technology *technology;
+	GSList *tech_drivers;
+	struct connman_technology_driver *driver;
 
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
@@ -1109,7 +1166,6 @@ void __connman_technology_remove_interface(enum connman_service_type type,
 		return;
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
 	case CONNMAN_SERVICE_TYPE_GPS:
@@ -1123,11 +1179,16 @@ void __connman_technology_remove_interface(enum connman_service_type type,
 
 	technology = technology_find(type);
 
-	if (technology == NULL || technology->driver == NULL)
+	if (technology == NULL)
 		return;
 
-	if (technology->driver->remove_interface)
-		technology->driver->remove_interface(technology, index);
+	for (tech_drivers = technology->driver_list; tech_drivers != NULL;
+	     tech_drivers = g_slist_next(tech_drivers)) {
+		driver = tech_drivers->data;
+
+		if(driver->remove_interface != NULL)
+			driver->remove_interface(technology, index);
+	}
 }
 
 int __connman_technology_add_device(struct connman_device *device)
@@ -1151,9 +1212,18 @@ int __connman_technology_add_device(struct connman_device *device)
 		return -ENXIO;
 	}
 
-	if (technology->enable_persistent &&
-					global_offlinemode == FALSE &&
-					technology->hardblocked == FALSE) {
+	__sync_synchronize();
+	if (technology->rfkill_driven == TRUE) {
+		if (technology->enabled == TRUE)
+			__connman_device_enable(device);
+		else
+			__connman_device_disable(device);
+
+		goto done;
+	}
+
+	if (technology->enable_persistent == TRUE &&
+					global_offlinemode == FALSE) {
 		int err = __connman_device_enable(device);
 		/*
 		 * connman_technology_add_device() calls __connman_device_enable()
@@ -1164,11 +1234,11 @@ int __connman_technology_add_device(struct connman_device *device)
 		if (err == -EALREADY)
 			__connman_technology_enabled(type);
 	}
-	/* if technology persistent state is offline or hardblocked */
-	if (technology->enable_persistent == FALSE ||
-					technology->hardblocked == TRUE)
+	/* if technology persistent state is offline */
+	if (technology->enable_persistent == FALSE)
 		__connman_device_disable(device);
 
+done:
 	technology->device_list = g_slist_prepend(technology->device_list,
 								device);
 
@@ -1200,17 +1270,36 @@ int __connman_technology_remove_device(struct connman_device *device)
 
 static void powered_changed(struct connman_technology *technology)
 {
-	connman_bool_t powered;
+	if (technology->dbus_registered == FALSE)
+		return;
+
+	if (technology->pending_reply != NULL) {
+		g_dbus_send_reply(connection,
+				technology->pending_reply, DBUS_TYPE_INVALID);
+		dbus_message_unref(technology->pending_reply);
+		technology->pending_reply = NULL;
+
+		g_source_remove(technology->pending_timeout);
+		technology->pending_timeout = 0;
+	}
 
 	__sync_synchronize();
-	if (technology->enabled >0)
-		powered = TRUE;
-	else
-		powered = FALSE;
-
 	connman_dbus_property_changed_basic(technology->path,
 			CONNMAN_TECHNOLOGY_INTERFACE, "Powered",
-			DBUS_TYPE_BOOLEAN, &powered);
+			DBUS_TYPE_BOOLEAN, &technology->enabled);
+}
+
+static int technology_enabled(struct connman_technology *technology)
+{
+	__sync_synchronize();
+	if (technology->enabled == TRUE)
+		return -EALREADY;
+
+	technology->enabled = TRUE;
+
+	powered_changed(technology);
+
+	return 0;
 }
 
 int __connman_technology_enabled(enum connman_service_type type)
@@ -1221,18 +1310,21 @@ int __connman_technology_enabled(enum connman_service_type type)
 	if (technology == NULL)
 		return -ENXIO;
 
-	if (__sync_fetch_and_add(&technology->enabled, 1) != 0)
+	if (technology->rfkill_driven == TRUE)
+		return 0;
+
+	return technology_enabled(technology);
+}
+
+static int technology_disabled(struct connman_technology *technology)
+{
+	__sync_synchronize();
+	if (technology->enabled == FALSE)
 		return -EALREADY;
 
-	powered_changed(technology);
+	technology->enabled = FALSE;
 
-	if (technology->pending_reply != NULL) {
-		g_dbus_send_reply(connection, technology->pending_reply, DBUS_TYPE_INVALID);
-		dbus_message_unref(technology->pending_reply);
-		g_source_remove(technology->pending_timeout);
-		technology->pending_reply = NULL;
-		technology->pending_timeout = 0;
-	}
+	powered_changed(technology);
 
 	return 0;
 }
@@ -1240,25 +1332,23 @@ int __connman_technology_enabled(enum connman_service_type type)
 int __connman_technology_disabled(enum connman_service_type type)
 {
 	struct connman_technology *technology;
+	GSList *list;
 
 	technology = technology_find(type);
 	if (technology == NULL)
 		return -ENXIO;
 
-	if (__sync_fetch_and_sub(&technology->enabled, 1) != 1)
-		return -EINPROGRESS;
+	if (technology->rfkill_driven == TRUE)
+		return 0;
 
-	if (technology->pending_reply != NULL) {
-		g_dbus_send_reply(connection, technology->pending_reply, DBUS_TYPE_INVALID);
-		dbus_message_unref(technology->pending_reply);
-		g_source_remove(technology->pending_timeout);
-		technology->pending_reply = NULL;
-		technology->pending_timeout = 0;
+	for (list = technology->device_list; list != NULL; list = list->next) {
+		struct connman_device *device = list->data;
+
+		if (connman_device_get_powered(device) == TRUE)
+			return 0;
 	}
 
-	powered_changed(technology);
-
-	return 0;
+	return technology_disabled(technology);
 }
 
 int __connman_technology_set_offlinemode(connman_bool_t offlinemode)
@@ -1287,10 +1377,10 @@ int __connman_technology_set_offlinemode(connman_bool_t offlinemode)
 		struct connman_technology *technology = list->data;
 
 		if (offlinemode)
-			err = technology_disable(technology, FALSE);
+			err = technology_disable(technology);
 
 		if (!offlinemode && technology->enable_persistent)
-			err = technology_enable(technology, FALSE);
+			err = technology_enable(technology);
 	}
 
 	if (err == 0 || err == -EINPROGRESS || err == -EALREADY) {
@@ -1320,41 +1410,72 @@ void __connman_technology_set_connected(enum connman_service_type type,
 			DBUS_TYPE_BOOLEAN, &connected);
 }
 
-static void technology_apply_hardblock_change(struct connman_technology *technology,
-						connman_bool_t hardblock)
+static connman_bool_t technology_apply_rfkill_change(struct connman_technology *technology,
+						connman_bool_t softblock,
+						connman_bool_t hardblock,
+						connman_bool_t new_rfkill)
 {
+	gboolean hardblock_changed = FALSE;
 	gboolean apply = TRUE;
 	GList *start, *list;
 
+	DBG("technology %p --> %d/%d vs %d/%d",
+			technology, softblock, hardblock,
+			technology->softblocked, technology->hardblocked);
+
 	if (technology->hardblocked == hardblock)
-		return;
+		goto softblock_change;
 
-	start = g_hash_table_get_values(rfkill_list);
-	for (list = start; list != NULL; list = list->next) {
-		struct connman_rfkill *rfkill = list->data;
+	if (!(new_rfkill == TRUE && hardblock == FALSE)) {
+		start = g_hash_table_get_values(rfkill_list);
 
-		if (rfkill->type != technology->type)
-			continue;
+		for (list = start; list != NULL; list = list->next) {
+			struct connman_rfkill *rfkill = list->data;
 
-		if (rfkill->hardblock != hardblock)
-			apply = FALSE;
+			if (rfkill->type != technology->type)
+				continue;
+
+			if (rfkill->hardblock != hardblock)
+				apply = FALSE;
+		}
+
+		g_list_free(start);
 	}
 
-	g_list_free(start);
-
 	if (apply == FALSE)
-		return;
+		goto softblock_change;
 
 	technology->hardblocked = hardblock;
+	hardblock_changed = TRUE;
 
-	if (hardblock == TRUE) {
-		DBG("%s is switched off.", get_name(technology->type));
-		technology_disable(technology, TRUE);
-		technology_dbus_unregister(technology);
-	} else {
-		technology_enable(technology, TRUE);
-		technology_dbus_register(technology);
+softblock_change:
+	if (apply == FALSE && technology->softblocked != softblock)
+		apply = TRUE;
+
+	if (apply == FALSE)
+		return technology->hardblocked;
+
+	technology->softblocked = softblock;
+
+	if (technology->hardblocked == TRUE ||
+					technology->softblocked == TRUE) {
+		if (technology_disabled(technology) != -EALREADY)
+			technology_affect_devices(technology, FALSE);
+	} else if (technology->hardblocked == FALSE &&
+					technology->softblocked == FALSE) {
+		if (technology_enabled(technology) != -EALREADY)
+			technology_affect_devices(technology, TRUE);
 	}
+
+	if (hardblock_changed == TRUE) {
+		if (technology->hardblocked == TRUE) {
+			DBG("%s is switched off.", get_name(technology->type));
+			technology_dbus_unregister(technology);
+		} else
+			technology_dbus_register(technology);
+	}
+
+	return technology->hardblocked;
 }
 
 int __connman_technology_add_rfkill(unsigned int index,
@@ -1389,22 +1510,25 @@ done:
 	if (technology == NULL)
 		return -ENXIO;
 
-	technology_apply_hardblock_change(technology, hardblock);
+	technology->rfkill_driven = TRUE;
+
+	/* If hardblocked, there is no need to handle softblocked state */
+	if (technology_apply_rfkill_change(technology,
+				softblock, hardblock, TRUE) == TRUE)
+		return 0;
 
 	/*
-	 * If Offline mode is on, we softblock the device if it isnt already.
-	 * If Offline mode is off, we rely on the persistent state of tech.
+	 * Depending on softblocked state we unblock/block according to
+	 * offlinemode and persistente state.
 	 */
-	if (global_offlinemode) {
-		if (!softblock)
-			return __connman_rfkill_block(type, TRUE);
-	} else {
-		if (technology->enable_persistent && softblock)
-			return __connman_rfkill_block(type, FALSE);
-		/* if technology persistent state is offline */
-		if (!technology->enable_persistent && !softblock)
-			return __connman_rfkill_block(type, TRUE);
-	}
+	if (technology->softblocked == TRUE &&
+				global_offlinemode == FALSE &&
+				technology->enable_persistent == TRUE)
+		return __connman_rfkill_block(type, FALSE);
+	else if (technology->softblocked == FALSE &&
+			(global_offlinemode == TRUE ||
+				technology->enable_persistent == FALSE))
+		return __connman_rfkill_block(type, TRUE);
 
 	return 0;
 }
@@ -1424,7 +1548,7 @@ int __connman_technology_update_rfkill(unsigned int index,
 		return -ENXIO;
 
 	if (rfkill->softblock == softblock &&
-		rfkill->hardblock == hardblock)
+				rfkill->hardblock == hardblock)
 		return 0;
 
 	rfkill->softblock = softblock;
@@ -1435,14 +1559,24 @@ int __connman_technology_update_rfkill(unsigned int index,
 	if (technology == NULL)
 		return -ENXIO;
 
-	technology_apply_hardblock_change(technology, hardblock);
+	/* If hardblocked, there is no need to handle softblocked state */
+	if (technology_apply_rfkill_change(technology,
+				softblock, hardblock, FALSE) == TRUE)
+		return 0;
 
-	if (!global_offlinemode) {
-		if (technology->enable_persistent && softblock)
-			return __connman_rfkill_block(type, FALSE);
-		if (!technology->enable_persistent && !softblock)
-			return __connman_rfkill_block(type, TRUE);
-	}
+	if (global_offlinemode == TRUE)
+		return 0;
+
+	/*
+	 * Depending on softblocked state we unblock/block according to
+	 * persistent state.
+	 */
+	if (technology->softblocked == TRUE &&
+				technology->enable_persistent == TRUE)
+		return __connman_rfkill_block(type, FALSE);
+	else if (technology->softblocked == FALSE &&
+				technology->enable_persistent == FALSE)
+		return __connman_rfkill_block(type, TRUE);
 
 	return 0;
 }
@@ -1465,6 +1599,9 @@ int __connman_technology_remove_rfkill(unsigned int index,
 	if (technology == NULL)
 		return -ENXIO;
 
+	technology_apply_rfkill_change(technology,
+		technology->softblocked, !technology->hardblocked, FALSE);
+
 	technology_put(technology);
 
 	return 0;
